/*
 * main.cc - startup code and application widget routines
 *
 * This file is part of the source distribution of 
 * Jafm - an icon based X11 desktop program 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 */

#include <qapp.h>
#include <qpmcache.h>
#include <qpaintd.h>
#include <qtstream.h>

#include "sfm.hh"
#include "iconview.hh"
#include "deskicon.hh"

#include <stdlib.h>
#include <X11/Xlib.h>
#include <stream.h>

static char copyr[] = "Jafm version 0.5 (alpha) Copyright 1997 Rupert Littlewood" ;

int main (int argc, char** argv)
{
  
  cout << "Jafm 0.5 (alpha release)\n"
       << "Jafm is freely distributable but comes with no warranty.\n" 
       << "See the file COPYING that came with this program for details.\n\n" ;
  cout.flush() ;

  Sfm app (argc, argv) ;

  // a few settings
#ifdef BIGCACHE
  QPixmapCache::setCacheLimit (4096) ;
#endif
  
  //  DeskIcon::theApp = &app ;
  Icon::load_rules () ; // probably shouldn't be called here

  QDir cur ;
  app.open (cur) ;
  app.openAppl ("Applications") ;
  return app.exec () ;
}


Sfm::Sfm (int argc, char **argv) 
  : QApplication (argc, argv) 
{
  winList = new QList<IconView> ;
  winIt = new QListIterator<IconView> (*winList) ;
  winIds = 0 ;
  diList = new QList<DeskIcon> ;
  diIt = new QListIterator<DeskIcon> (*diList) ;
  stackList = 0 ;
  retrieveSession () ;
} 


void Sfm::dragPointAt (const QPoint &p, const QList<DeskIcon> &icList)
{
  /*
   * we have been informed that something is being dragged
   * and the pointer is at p, in global coords
   */
  
  // we need to use Xlib directly here.

  Window rootWin ;
  Window childWin ;
  QPixmap pm ; //kludge
  int k, j ;
  // check if we have examined the window tree
  if (!stackList) {
    // we need to know the parent id
    winIt->toFirst () ;
    Window parWin ;
    if (!winIds) delete[] winIds ;
    winIds = new WId[winList->count()] ;
    k = 0 ;
    for ( ; winIt->current() ; ++(*winIt)) {
      Window innerWin = winIt->current()->winId() ;
      do {
	if (XQueryTree(winIt->current()->x11Display(), innerWin,
		       &rootWin, &parWin, &stackList, &stackLen)) {
	  XFree (stackList) ;
	  winIds[k] = innerWin ; 
	  innerWin = parWin ;
	} else {
	  debug ("XQueryTree failed") ;
	  exit (1) ;
	}
      } while (rootWin != parWin) ; // window managers vary - nested parents
      k++ ;
    }
    
    if ( ! XQueryTree (QPaintDevice::x__Display(), DefaultRootWindow(QPaintDevice::x__Display()),
		       &rootWin, &childWin, 
		       &stackList, &stackLen)) {
      debug ("XQueryTree failed") ;
      exit (1) ;
    }
  }
  
  // this algorithm is poor, we count on not having many windows
  diIt->toFirst () ;
  for (j = stackLen-1 ; j >= 0 ; --j) {
    HANDLE h = stackList[j] ;
    // debug ("testing handle %x",h) ;
    // is it in a window ?
    winIt->toFirst () ;
    k = 0 ;
    for ( ; winIt->current() ; ++(*winIt)) {
      //      debug ("       Is it %x or %x", winIt->current()->winId(), winIds[k]) ;
      if (winIds[k++] == h) {
	if (winIt->current()->geometry().contains (p)) {
	  emit (dragDropHighlight (winIt->current()->winId(), p, icList )) ;
	  return ;
	}
      }
    }
    // is it in a deskicon
    diIt->toFirst () ;
    for ( ; diIt->current() ; ++(*diIt)) {
      //    debug ("is %x equal to %x",diIt->current()->winId(), h) ;
      if (diIt->current()->winId() == h) {
	
	//	if (diIt->current()->geometry().contains (p)) {
	// above line fails? gives to large a geometry
	int w = diIt->current()->get_icon()->width() ;
	int ht = diIt->current()->get_icon()->height() ;
	  
	if (diIt->current()->x()<p.x() && diIt->current()->x()+w>p.x() 
	    && diIt->current()->y()<p.y() && diIt->current()->y()+ht>p.y()) {
	  emit (dragDropHighlight (h, p, icList)) ;
	  return ;	
	}
      }
    }
  }
  // nowhere!
  emit (dragDropHighlight (0, p, icList)) ; // allows windows to unhighlight if nec
}

void Sfm::dragDropRelease (const QPoint &p, const QList<DeskIcon> &icList) 
{
  /*
   * we have been informed that something has been dropped
   * at p, in global coords
   */
  // we need to use Xlib directly here.

  // this algorithm is poor, we count on not having many windows
  int j, k ;
  diIt->toFirst () ;
  for (j = stackLen-1 ; j >= 0 ; --j) {
    HANDLE h = stackList[j] ;
    // debug ("testing handle %x",h) ;
    // is it in a window ?
    winIt->toFirst () ;
    k = 0 ;
    for ( ; winIt->current() ; ++(*winIt)) {
      //      debug ("       Is it %x or %x", winIt->current()->winId(), winIds[k]) ;
      if (winIds[k++] == h) {
	if (winIt->current()->geometry().contains (p)) {
	  emit (drag_drop_happened (winIt->current()->winId(), p, icList)) ;
	  XFree (stackList) ; stackList = 0 ;
	  return ;
	}
      }
    }
    // is it in a deskicon
    diIt->toFirst () ;
    for ( ; diIt->current() ; ++(*diIt)) {
      if (diIt->current()->winId() == h) {
	//	if (diIt->current()->geometry().contains (p)) {
	int w = diIt->current()->get_icon()->width() ;
	int ht = diIt->current()->get_icon()->height() ;
	  
	if (diIt->current()->x()<p.x() && diIt->current()->x()+w>p.x() 
	    && diIt->current()->y()<p.y() && diIt->current()->y()+ht>p.y()) {
	  // debug ("In deskicon") ;
	  emit (drag_drop_happened (h, p, icList )) ;
	  XFree (stackList) ; stackList = 0 ;
	  return ;
	}
      }
    }
  }
  
  // on desktop!
    // otherwise, dropped on desktop?
    
  // debug ("desktop drop") ;
  
  QListIterator<DeskIcon> itt (icList) ;
  for (; itt.current() ; ++itt) {
    diList->append (new DeskIcon (itt.current(), false)) ;
    diList->current()->show () ;
  }

  XFree (stackList) ; stackList = 0 ;
}


void Sfm::slot_file_change (const QFileInfo &f,const QFileInfo &newf) 
{
  /* 
   * the file with (abs) name f has changed to newf,
   * if it exists on desktop change it.
   * Pass message on to others
   */

  emit broadcast_file_change (f, newf) ;
}



bool Sfm::retrieveSession ()
{
  /*
   * set up windows, icons as the user last had them
   * if not possible return false, else return true
   */

  return true ;

  // is there a home
  const char *home = getenv ("HOME") ;
  if (strlen  (home) == 0) {
    debug ("No home directory!") ;
    return false ;
  }
  QDir::setCurrent (home) ; 
  
  //  is there an .jafm dir?
  if (! QDir(".jafm").exists()) {
    debug ("No .jafm dir in %s, creating...", home) ;
    QDir d ;
    d.mkdir (".jafm") ;
    return false ;
  }
  
  // is there a saved session?
  QDir::setCurrent (".jafm") ;
  QFile f (".session") ;
  if (f.exists()) {
    debug ("Have .session file") ;
    //   QTextStream 

    return true ;

  }
  debug ("No .session file.") ;
  return false ;
}
    
  
void Sfm::slot_refresh ()
{
  /* An icondesk has indicated that a change has
   * occurred in its window - we need to immediately
   * check the other windows to keep in sync
   */

  // send messages to all the iconviews
  
  winIt->toFirst() ;
  for ( ; winIt->current() ; ++*winIt) 
    winIt->current()->refresh() ;

}

